<TeXmacs|1.0.2.9>

<style|mmxdoc>

<\body>
  <tmdoc-title|The <value|mmx> type system>

  The <value|mmx> type system is based on a few general principles:

  <\itemize>
    <item>The declaration of a symbol is seen as a notification of the
    <em|ability> to carry out a certain task under certain circumstances.
    Several solutions may be specified for the same task under different
    circumstances (overloading and natural extensability).

    <item>Ambiguous expressions systematically carry types and <value|mmx>
    contains a general mechanism for computations with ambiguous types.

    <item>As a counterpart to the powerful constructs which are provided by
    the type system, the user has some responsability of using it in the
    appropriate way. When used inappropriately, some of the type checking
    algorithms may not terminate or abort prematurely.

    <item>The type system should be as ``independent'' as possible from the
    rest of the language.
  </itemize>

  Concretely speaking, <value|mmx> types are formed from

  <\description>
    <item*|Atomic symbols.>Examples: <with|mode|math|<type|String>>,
    <with|mode|math|<type|Int>>, etc.

    <item*|Type constructors.>Example: <with|mode|math|<type|Polynomial>(<type|Int>),<type|Int>\<rightarrow\><type|String>>,
    <abbr|etc.>. Type constructors may also use instances, like in
    <with|mode|math|<type|Array>(<type|Int>,3)>.

    <item*|Logical constructors.>Examples:
    <with|mode|math|<type|Int><value|andt><type|String>> (overloaded
    <with|mode|math|<type|Int>> and <with|mode|math|<type|String>>),
    <with|mode|math|<value|forallt>X,X\<rightarrow\>X> (template type).
  </description>

  Let us explain these ideas in a bit more detail.

  <paragraph|Logical supertypes>

  In classical type systems, each variable <with|mode|math|x> has a type
  <with|mode|math|T> (written <with|mode|math|x:T>). In <value|mmx>, one
  should rather see the fact that ``the context provides a symbol
  <with|mode|math|x> of type <with|mode|math|T>'' (written
  <with|mode|math|x:T>) as a <em|statement>. Such statements may be combined
  in the language of first order logic. Now logical supertypes are new types
  which correspond to such combinations. For instance, the type
  <with|mode|math|T<value|andt>U> is the type such that for all symbols
  <with|mode|math|x>, we have

  <\equation*>
    x:T<value|andt>U\<Longleftrightarrow\>x:T\<wedge\>x:U.
  </equation*>

  In other words, if <with|mode|math|x:T<value|andt>U> in a given context,
  then <with|mode|math|x> may both be used as an object of type
  <with|mode|math|T> or as an object of type <with|mode|math|U>, <abbr|i.e.>
  <with|mode|math|x> has been overloaded, or the result of an ambiguous
  expression. The following logical type constructors are currently
  implemented in the type system (although only
  <no-break><with|mode|math|<value|andt>> and
  <with|mode|math|<value|forallt>> are currently available in the
  interpreter):

  <\itemize>
    <item><with|mode|math|x:T<value|andt>U\<Leftrightarrow\>x:T\<wedge\>x:U>
    for overloading.

    <item><with|mode|math|x:T<value|ort>U\<Leftrightarrow\>x:T\<vee\>x:U> for
    union-like types.

    <item><with|mode|math|x:<value|forallt>a,T(a)\<Leftrightarrow\>\<forall\>a,x:T(a)>
    for template types. For instance, if <with|mode|math|f:<value|forallt>T,T\<rightarrow\>T>,
    then for any type <no-break><with|mode|math|T>, we may use
    <with|mode|math|f> as a function from <with|mode|math|T> to
    <with|mode|math|T>.

    <item><with|mode|math|x:<value|existst>a,T(a)\<Leftrightarrow\>\<exists\>a,x:T(a)>
    for generic types. For instance, a routine to find the roots of a
    polynomial might have the type <with|mode|math|<value|forallt>T,<value|existst>U,<type|Polynomial>(T)\<rightarrow\><type|List>(U)>.
  </itemize>

  In the future, we also plan to add conditional types
  <with|mode|math|R<value|implt>T>, with <with|mode|math|x:R<value|implt>T\<Leftrightarrow\>(R\<Rightarrow\>x:T)>.
  The relations <with|mode|math|R> should be defined using first order logic
  axioms (and possibly refer to predicate routines). For instance, this would
  permit to use types like <with|mode|math|<value|forallt>T,Ring(T)<value|implt>T\<rightarrow\>T>.

  <paragraph|Extensibility and exceptions>

  The advantage of logical supertypes is that it is possible to define
  functions which are only valid under very precise circumstances, which are
  modelized using first order logic. This makes the system very extensible
  and customizable.

  Indeed, assume for instance that you have functions <with|mode|math|f> and
  <with|mode|math|g>, where <with|mode|math|g> depends on <with|mode|math|f>
  and where <with|mode|math|f> is valid under certain circumstances. Then it
  may be that someone makes a new implementation of <with|mode|math|f> under
  another circumstance. Then it becomes automatically possible to apply
  <with|mode|math|g> in this new circumstance.

  It may also be that the original implementation of <with|mode|math|f> is
  valid under very broad hypothesis, but also very unefficient. Then a new
  user may reimplement <with|mode|math|f> under more particular hypothesis
  later. Here <value|mmx> uses another general principle: when applying an
  overloaded function to some type, then the most particular implementation
  wins. In other words, if <with|mode|math|f:T\<rightarrow\>()<value|andt>U\<rightarrow\>()>
  is applied to <no-break><with|mode|math|x:T>, and the typechecker can prove
  <with|mode|math|x:T\<rightarrow\>x:U>, but not
  <with|mode|math|x:U\<rightarrow\>x:T>, then
  <with|mode|math|f:T\<rightarrow\>()> is applied. It should be noticed that
  this principle makes it unnecessary to introduce negation in the logical
  type system. In fact, negation would be harmful, because it breaks the
  extensibility.

  Another major advantange of the logical supertype approach is that it
  allows for less hierarchic design of big projects. For example, consider
  the problem where you have trees with <with|mode|math|n> types of possible
  nodes and that you need to define <with|mode|math|k> types of actions on
  trees which depend on the nodes, but where many implementations can be
  grouped on either node type or action type. In <name|C++> this can
  essentially be done in two ways: either you declare the tree class as a
  structure with <with|mode|math|k> virtual functions, which makes it
  possible to regroup the implementations by node types, or you declare a
  simple tree class with enumeration nodes, and you regroup the
  implementations by their action type. In either case, one design style is
  priviledged: dispatching on node type or dispatching on action type. In
  <value|mmx>, it is naturally possible to mix both styles.

  <paragraph|Feasability>

  Of course, it is nice to design a very powerful type system, but it is also
  nice if one can implement <no-break>it. In the most general setting of
  first order logic, this is quite hopeless. Nevertheless, only a tiny part
  of the first order logic is used in practice, even though it may be hard to
  indentify this part. Our approach is to progressively extend the strength
  of the implementation of the type system and accept the fact that it will
  not do everything. This should make certain types of uses perfectly
  reliable, but it will remain up to the user not to abuse the type system.

  In order to make this philosophy fully functional, it is nice to make the
  type system as independent as possible from the rest of the language, so
  that it might actually be carried out by a dedicated systeem, like a
  theorem proving system. In <value|mmx> the implementation of the type
  system is in the directory <verbatim|src/logic> and you may test it using
  <verbatim|mmx-convert>. Typically the type system should provide methods
  for type conversion, typing function applications and controlling implicit
  conversions.

  At the moment, the type system is able to handle pure predicate logic
  (restricted to <with|mode|math|<value|andt>> and
  <with|mode|math|<value|ort>>) well. It can also deal with types of the form

  <\equation*>
    <value|forallt>a<rsub|1>,\<ldots\>,<value|forallt>a<rsub|k>,<value|existst>b<rsub|1>,\<ldots\>,<value|existst>b<rsub|l>,T(a<rsub|1>,\<ldots\>,a<rsub|k>)\<rightarrow\>U(a<rsub|1>,\<ldots\>,a<rsub|k>,b<rsub|1>,\<ldots\>,b<rsub|l>)
  </equation*>

  in a systematic way, and conjunctions of such types. Many easy other types
  can also be dealt with, though not systematically. A current limitation is
  that when you have a method

  <\equation*>
    f:<value|forallt>X,(X,X)\<rightarrow\>()
  </equation*>

  and an implicit converter <with|mode|math|T\<rightsquigarrow\>U>, then
  <with|mode|math|f> can not yet be applied to <with|mode|math|(T,U)>.

  <paragraph|Semantics>

  The precise semantics of logical supertypes is still under development. One
  has to decide in particular if one wants to restrict the application of
  quantifiers to atomic supertypes (i.e. <with|mode|math|<type|String>>,
  <with|mode|math|<type|Polynomial>(<type|Int>)> or
  <with|mode|math|<type|List>(<type|String><value|andt><type|Int>)> are OK,
  but not <with|mode|math|<type|String><value|andt><type|Int>>). Indeed,
  consider the function

  <\equation*>
    f:<value|forallt>X,(X,X)\<rightarrow\>()
  </equation*>

  If one may take <with|mode|math|X=T<value|ort>U>, then, using the
  convertions <with|mode|math|T\<rightsquigarrow\>T<value|ort>U> and
  <with|mode|math|U\<rightsquigarrow\>T<value|ort>U>, it becomes possible to
  applie <with|mode|math|f> to <with|mode|math|(T,U)> since this is the case
  for all <with|mode|math|T> and <with|mode|math|U>, we obtain an implicit
  converter

  <\equation>
    <label|abuse-conv><value|forallt>X,(X,X)\<rightarrow\>()<space|1.2spc>\<rightsquigarrow\><space|1.2spc><value|forallt>T,<value|forallt>U,(T,U)\<rightarrow\>()
  </equation>

  Here one hits the limitations of the approach of not requiring any
  additional information on types, like a category in <name|Aldor>. Indeed,
  if one writes a routine like

  <\equation*>
    square:<value|forallt>X,X\<rightarrow\>X;x\<mapsto\>x\<times\>x
  </equation*>

  then it is not really necessary to know the type of <with|mode|math|X>,
  since it is very likely that the user will only apply it in the case when
  there is a multiplication defined on <with|mode|math|X>. However, if one
  has multiplications on <with|mode|math|T> and <with|mode|math|U>, then it
  is highly unlikely to have a multiplication on
  <with|mode|math|T<value|ort>U>, so the implicit conversion
  <no-break>(<reference|abuse-conv>) is generally abusive.

  <tmdoc-copyright|2003|Joris van der Hoeven>

  <tmdoc-license|Permission is granted to copy, distribute and/or modify this
  document under the terms of the <hlink|GNU General Public
  License|http://www.gnu.org/licenses/gpl.txt>. If you don't have this file,
  write to the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
  Boston, MA 02111-1307, USA.>
</body>

<\initial>
  <\collection>
    <associate|page-even|30mm>
    <associate|page-reduce-bot|15mm>
    <associate|page-reduce-right|25mm>
    <associate|page-reduce-left|25mm>
    <associate|page-top|30mm>
    <associate|page-right|30mm>
    <associate|par-width|150mm>
    <associate|page-odd|30mm>
    <associate|page-bot|30mm>
    <associate|page-reduce-top|15mm>
  </collection>
</initial>

<\references>
  <\collection>
    <associate|abuse-conv|<tuple|1|?>>
    <associate|toc-1|<tuple|<uninit>|?>>
    <associate|toc-2|<tuple|<uninit>|?>>
    <associate|toc-3|<tuple|<uninit>|?>>
    <associate|toc-4|<tuple|<uninit>|?>>
  </collection>
</references>

<\auxiliary>
  <\collection>
    <\associate|toc>
      <with|par-left|<quote|6fn>|font-size|<quote|0.84>|Logical
      supertypes<value|toc-dots><pageref|toc-1>>
    </associate>
  </collection>
</auxiliary>