
/******************************************************************************
* MODULE     : evaluator.hh
* DESCRIPTION: the mathemagix evaluator
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __EVALUATOR_HH
#define __EVALUATOR_HH
#include <symbol.hh>
#include <table.hh>
#include <xtable.hh>
#include <list.hh>
#include <value.hh>
#include <keyword.hh>
#include <extern.hh>
namespace mmx {

extern bool debug_output;

#ifdef DEBUG_ON
#define CHECK_SYNC(kw) \
  if (!is_func (car (trace), kw)) { \
    error (car (trace), "Lost synchronization"); \
    return; \
  }
#define CHECK_SYNC2(kw1,kw2) \
  if (!(is_func (car (trace), kw1) || (is_func (car (trace), kw2)))) { \
    error (car (trace), "Lost synchronization"); \
    return; \
  }
#define FAST_CONTINUE(with)
#define SHOW_ENV(action,t)
#define SHOW_ENV_VAR(action,t) \
  cout << hrule << action << " " << env->local_level() << ", "; \
  if (nil (trace)) cout << "NIL" << lf; \
  else cout << car(trace) << lf; \
  dump (t);
#else
#define CHECK_SYNC(kw)
#define CHECK_SYNC2(kw1,kw2)
#define FAST_CONTINUE(with) with()
#define SHOW_ENV(action,t)
#endif

class value;
class evaluator_rep;
typedef void (evaluator_rep::*mmx_keyword) ();

class evaluator {
protected:
  evaluator_rep* rep;

public:
  inline evaluator ();
  inline evaluator (evaluator_rep* rep2): rep (rep2) {}
  inline friend evaluator as_evaluator (generic g) {
    if (nil (g)) return evaluator ();
    if (gentype (g) != EVALUATOR_TYPE) type_mismatch (g, "evaluator");
    return evaluator ((evaluator_rep*) g.rep); }
  inline operator generic ();
  inline evaluator_rep* operator -> () { return rep; }

  friend class evaluator_rep;
  friend ostream& operator << (ostream& out, evaluator con);
};

class evaluator_rep: generic_rep {
public:
  gentable env;   // the current environment
  generic  cur;   // next instruction to be interpreted
  list     rem;   // parameters for cur and remaining instructions
  list     trace; // original instructions being executed

protected:
  generic apply_prog (generic fun_type, generic arg_type);

public:
  evaluator_rep ();

  int  Type ();
  void Print (ostream& out);
  
  // declarations
  void declare (keyword kw);
  void declare (symbol var, value val);
  void global_declare (symbol var, value val);

  // evaluation
  generic loop ();
  generic evaluate (generic g);
  void    error (generic where, string message);

  void evaluate_arg (generic arg, keyword next);
  void evaluate_args (list args, keyword next);
  void evaluate_args (keyword next);
  void display_one (value v);
  void display (value v);

  // changing environments
  inline void start_local_env (gentable t);
  inline void end_local_env ();
  inline void change_env (gentable t);

  // keywords
  void kw_compound_1 ();
  void kw_compound_2 ();
  void kw_compound_3 ();
  void kw_eval_args_1 ();

  void kw_conv_aliases_1 ();
  void kw_conv_aliases_2 ();
  void kw_conv_ambiguous_1 ();
  void kw_conv_apply_1 ();
  void kw_conv_apply_2 ();
  void kw_conv_apply_3 ();
  void kw_conv_cross_1 ();
  void kw_conv_cross_2 ();
  void kw_conv_forall_1 ();
  void kw_conv_ignore_1 ();
  void kw_conv_lambda_1 ();
  void kw_conv_singlet_1 ();
  void kw_conv_singlet_2 ();
  void kw_conv_template_1 ();
  void kw_conv_template_2 ();
  void kw_conv_template_3 ();
  void kw_conv_unalias_1 ();
  void kw_conv_unalias_2 ();
  void kw_conv_unand_1 ();
  void kw_conv_unand_2 ();
  void kw_conv_unand_3 ();
  void kw_conv_uncross_1 ();
  void kw_conv_uncross_2 ();
  void kw_conv_vcross_1 ();
  void kw_conv_vcross_2 ();

  void kw_assign_1 ();
  void kw_assign_2 ();
  void kw_assign_3 ();
  void kw_begin_1 ();
  void kw_begin_2 ();
  void kw_break_1 ();
  void kw_continue_1 ();
  void kw_cross_1 ();
  void kw_cross_2 ();
  void kw_case_1 ();
  void kw_case_2 ();
  void kw_case_3 ();
  void kw_case_4 ();
  void kw_define_1 ();
  void kw_define_2 ();
  void kw_define_3 ();
  void kw_for_1 ();
  void kw_forallval_1 ();
  void kw_forallval_2 ();
  void kw_forin_1 ();
  void kw_forin_2 ();
  void kw_forin_3 ();
  void kw_forin_4 ();
  void kw_forin_5 ();
  void kw_forin_6 ();
  void kw_function_1 ();
  void kw_function_2 ();
  void kw_generate_1 ();
  void kw_generate_2 ();
  void kw_generate_3 ();
  void kw_if_1 ();
  void kw_if_2 ();
  void kw_join_1 ();
  void kw_join_2 ();
  void kw_join_3 ();
  void kw_lambda_1 ();
  void kw_lambda_2 ();
  void kw_load_1 ();
  void kw_load_2 ();
  void kw_local_1 ();
  void kw_local_2 ();
  void kw_loop_1 ();
  void kw_loop_2 ();
  void kw_loop_3 ();
  void kw_loop_4 ();
  void kw_quantor_1 ();
  void kw_quantor_2 ();
  void kw_retrieve_1 ();
  void kw_retrieve_2 ();
  void kw_retrieve_3 ();
  void kw_return_1 ();
  void kw_return_2 ();
  void kw_split_1 ();
  void kw_split_2 ();
  void kw_step_1 ();
  void kw_transtype_1 ();
  void kw_transtype_2 ();
  void kw_tuple_1 ();
  void kw_tuple_2 ();
  void kw_until_1 ();
  void kw_use_1 ();
  void kw_use_2 ();
  void kw_while_1 ();
  void kw_yield_1 ();
  void kw_yield_2 ();

  void lib_print_any_1 ();
  void lib_print_class_1 ();
  void lib_print_class_2 ();
  void lib_print_class_3 ();
  void lib_print_cross_1 ();
  void lib_print_cross_2 ();
  void lib_print_cross_3 ();
  void lib_print_tuple_1 ();
  void lib_print_tuple_2 ();
  void lib_print_tuple_3 ();
  void lib_print_alias_1 ();
  void lib_print_code_1 ();

  friend class evaluator;
};

inline evaluator::evaluator (): rep (new evaluator_rep ()) {}
inline evaluator::operator generic () { return generic (rep); }

inline void
evaluator_rep::start_local_env (gentable t) {
  SHOW_ENV ("Start environment", t);
  env= xtable (env, t);
}

inline void
evaluator_rep::end_local_env () {
  env= as_xtable (env)->base_t;
  SHOW_ENV ("End environment", table ());
}

inline void
evaluator_rep::change_env (gentable t) {
  env= t;
  SHOW_ENV ("Change environment", table ());
}

extern evaluator* mmeval;
generic mmx_evaluate (generic g);
void mmx_display (value v);

}
#endif // __EVALUATOR_HH
