
/******************************************************************************
* MODULE     : ext_convert.cc
* DESCRIPTION: Conversion routines between Mathemagix and C++
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <single.hh>
namespace mmx {

/******************************************************************************
* Matching types
******************************************************************************/

static int match_counter= 0;

bool
match_type (generic what, generic in, table t) {
  // cout << what << " --- " << in << lf;
  if (what == in) return true;
  if (t->contains (in)) {
    if (!nil (t[in])) return match_type (what, t[in], t);
    t(in)= what;
    return true;
  }
  if (is_func (what, MMX_FORALL, 2)) {
    generic what_var= what[1];
    if (is_func (what_var, MMX_TYPE, 2)) what_var= what_var[1];
    generic sub= substitute (what[2], what_var, single (match_counter++));
    return match_type (sub, in, t);
  }
  if (is_func (in, MMX_FORALL, 2)) {
    generic in_var= in[1];
    if (is_func (in_var, MMX_TYPE, 2)) in_var= in_var[1];
    t= copy (t);
    t(in_var)= generic ();
    return match_type (what, in[2], t);
  }
  if (is_func (in, MMX_AND)) {
    int i, n= N(in);
    for (i=1; i<n; i++)
      if (match_type (what, in[i], t)) return true;
  }
  if (is_list (what) && is_list (in) && (!nil (what)) && (!nil (in))) {
    list what_l= as_list (what);
    list in_l  = as_list (in);
    while ((!nil (what_l)) && (!nil (in_l))) {
      if (!match_type (car (what_l), car (in_l), t)) return false;
      what_l= cdr (what_l);
      in_l= cdr (in_l);
    }
    return nil (what_l) && nil (in_l);
  }
  return false;
}

generic
get_value (generic what, list l) {
  generic var1 = get_decl_var  (what);
  generic type1= get_decl_type (what);
  while (!nil (l)) {
    generic var2 = get_decl_var  (car (l));
    generic type2= get_decl_type (car (l));
    if ((var1 == var2) && match_type (type1, type2))
      return get_decl_val (car (l));
    l= cdr (l);
  }
  return generic ();
}

/******************************************************************************
* Compute the C/C++ name of a Mathemagix type. The table t contains
* already computed conversions for a finite number of additional types,
* usually template arguments
******************************************************************************/

generic
extender::type_as_ccpp (generic g, table t) {
  // search the C/C++ name of a Mathemagix type
  // some additional conversion rules may be specified in the table
  if (t -> contains (g)) return t[g];
  else if (ex_type_conv -> contains (g)) return ex_type_conv [g];
  // else if (is_func (g, MMX_CROSS, 0)) return MMX_VOID;
  else if (is_func (g, MMX_TUPLETYPE, 1) || is_func (g, MMX_VTUPLETYPE, 1)) {
    generic tup= Scope (symbol ("mmx"), symbol ("glued_tuple"));
    return list (tup, type_as_ccpp (g[1], t), genid);
  }
  else if (is_list (g)) {
    list l= as_list (g), r;
    while (!nil (l)) {
      r= cons (type_as_ccpp (car (l), t), r);
      l= cdr (l);
    }
    return reverse (r);
  }
  else if (g == MMX_OSTREAM) return Alias (g);
  else if (g == MMX_CLASS) return gentype;
  else if ((g == MMX_VOID) ||
	   (g == MMX_BOOL) ||
	   (g == MMX_INT) ||
	   (g == MMX_LITERAL_INT) ||
	   (g == MMX_LITERAL_FLOAT) ||
	   (g == MMX_LITERAL_STRING) ||
	   (g == MMX_ALIAS) ||
	   (g == MMX_CROSS) ||
	   (g == MMX_VCROSS) ||
	   (g == MMX_FUNCTION) ||
	   (g == CCPP_CONST) ||
	   (g == CCPP_VIRTUAL) ||
	   (g == geninst ()))
    return g;
  else {
    int i;
    for (i=1; i<=max_type_args; i++)
      if (g == geninst (i))
	return g;

    warning (g, "Type could not be translated");
    return g;
  }
}

list
extender::type_as_ccpp (list l, table t) {
  if (nil (l)) return l;
  return cons (type_as_ccpp (car (l), t), type_as_ccpp (cdr (l), t));
}

/******************************************************************************
* Convert a Mathemagix declaration into a C++ declaration. The gen_flag
* tells whether the declaration is a method of a generic type structure.
******************************************************************************/

generic
extender::decl_as_ccpp (generic g, bool gen_flag) {
  // cout << "Declaration as c++ " << g << lf;
  if (!(is_func (g, MMX_DEFINE, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	is_symbol (g[1][1]) &&
	is_symbol (g[2])))
    error (g, "Assertion failed in decl_as_ccpp");
  else if (!gen_flag)
    return list (g[1][0], g[2], type_as_ccpp (g[1][2]));
  else return list (g[1][0], genmethod (g[2]), type_as_ccpp (g[1][2]));
  return g;
}

list
extender::decl_as_ccpp (list l, bool gen_flag) {
  if (nil (l)) return l;
  else return cons (decl_as_ccpp (car (l), gen_flag),
		    decl_as_ccpp (cdr (l), gen_flag));
}

/******************************************************************************
* Viewing a Mathemagix generic as a C++ type
******************************************************************************/

generic
extender::view_as_ccpp (generic g, generic mmx_type, table t) {
  generic type= type_as_ccpp (mmx_type, t);
  generic cv  = mmx_specialize ("as_cpp", type, genid);
  generic expr= g;
  if (mmx_type == MMX_LITERAL_INT) cv= mmx_symbol ("as_literal_contents");
  if (mmx_type == MMX_LITERAL_FLOAT) cv= mmx_symbol ("as_literal_contents");
  if (mmx_type == MMX_LITERAL_STRING) cv= mmx_symbol ("as_literal_contents");
  if (is_func (mmx_type, MMX_ALIAS, 1))
    if ((mmx_type[1] == MMX_CLASS) ||
	(mmx_type[1] == MMX_OSTREAM) ||
	(mmx_type[1] == MMX_LITERAL_INT) ||
	(mmx_type[1] == MMX_LITERAL_FLOAT) ||
	(mmx_type[1] == MMX_LITERAL_STRING))
      error (mmx_type, "Unsupported alias type");
  if (is_func (mmx_type, MMX_TUPLETYPE, 1)) {
    cv= mmx_specialize ("glued_tuple", type[1], genid);
    if (is_func (expr, MMX_ACCESS, 2)) // for security
      expr= expr[1]; // temporary hack to get arguments right
  }
  if (is_func (mmx_type, MMX_VTUPLETYPE, 1))
    cv= mmx_specialize ("glued_tuple", type[1], genid);
  return list (cv, expr);
}

/******************************************************************************
* Viewing a C++ type as a Mathemagix generic
******************************************************************************/

generic
extender::view_as_mmx (generic expr, generic mmx_type, table t) {
  generic type= type_as_ccpp (mmx_type, t);
  generic cv  = mmx_specialize ("as_mmx", type, genid);
  if (mmx_type == MMX_LITERAL_INT) cv= mmx_symbol ("literal_int");
  if (mmx_type == MMX_LITERAL_FLOAT) cv= mmx_symbol ("literal_float");
  if (mmx_type == MMX_LITERAL_STRING) cv= mmx_symbol ("literal_string");
  if (is_func (mmx_type, MMX_ALIAS, 1))
    if ((mmx_type[1] == MMX_CLASS) ||
	(mmx_type[1] == MMX_OSTREAM) ||
	(mmx_type[1] == MMX_LITERAL_INT) ||
	(mmx_type[1] == MMX_LITERAL_FLOAT) ||
	(mmx_type[1] == MMX_LITERAL_STRING))
      error (mmx_type, "Unsupported alias type");
  if (is_func (mmx_type, MMX_TUPLETYPE, 1))
    cv= mmx_specialize ("as_generic", type[1], genid);
  if (is_func (mmx_type, MMX_VTUPLETYPE, 1))
    cv= mmx_specialize ("as_generic", type[1], genid);
  return list (cv, expr);
}

list
extender::view_as_mmx (list exprs, list mmx_types, table t) {
  if (nil (exprs) || nil (mmx_types)) return list ();
  return cons (view_as_mmx (car (exprs), car (mmx_types), t),
	       view_as_mmx (cdr (exprs), cdr (mmx_types), t));
}

} // namespace mmx
