
/******************************************************************************
* MODULE     : ext_glue.cc
* DESCRIPTION: Generate glue with Mathemagix
* COPYRIGHT  : (C) 2004  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <extend.hh>
#include <ccpp.hh>
#include <single.hh>
namespace mmx {

/******************************************************************************
* Building Mathemagix types in glue initialization
******************************************************************************/

static int
find_in_bindings (generic g, list l, int i=1) {
  if (nil (l)) return -1;
  if (is_func (car (l), MMX_TYPE, 2) && (car (l) [1] == g)) return i;
  return find_in_bindings (g, cdr (l), i+1);
}

bool
extender::match_mmx_type (generic expr, generic defn, list bindings) {
  // cerr << expr << " == " << defn << lf;
  if (expr == defn) return true;
  int i= find_in_bindings (expr, bindings);
  if (i != -1) return match_mmx_type (bindings[i-1][2], defn, list ());
  if ((!is_list (expr)) || nil (expr)) return false;
  if ((!is_list (defn)) || nil (defn)) return false;
  return
    match_mmx_type (car (expr), car (defn), bindings) &&
    match_mmx_type (cdr (expr), cdr (defn), bindings);
}

int
extender::resolve_mmx_type (generic expr, list l, int i, list bindings) {
  // FIXME: this should really be done using the converter from logic
  // FIXME: moreover the searching is only quadratic in time
  if (nil (l)) return -1;
  generic defn= car (l);
  if (is_func (defn, MMX_DEFINE, 2)) defn= defn[1];
  if (is_func (defn, MMX_TYPE, 2)) {
    if (is_func (defn[2], MMX_FUNCTION, 2)) {
      if (is_func (defn[2][1], MMX_CROSS))
	defn= cons (defn[1], cdr (as_list (defn[2][1])));
    }
    else defn= defn[1];
  }
  if (match_mmx_type (expr, defn, bindings)) return i;
  return resolve_mmx_type (expr, cdr (l), i+1, bindings);
}

generic
extender::build_mmx_type_sub (generic g, list bindings) {
  if (g == MMX_CLASS) return mmx_symbol ("ID_CLASS");
  if (g == MMX_VOID) return mmx_symbol ("ID_VOID");
  if (g == MMX_BOOL) return mmx_symbol ("ID_BOOL");
  if (g == MMX_INT) return mmx_symbol ("ID_INT");
  if (g == MMX_OSTREAM) return mmx_symbol ("ID_OSTREAM");
  if (g == MMX_LITERAL_INT) return mmx_symbol ("ID_LITERAL_INT");
  if (g == MMX_LITERAL_FLOAT) return mmx_symbol ("ID_LITERAL_FLOAT");
  if (g == MMX_LITERAL_STRING) return mmx_symbol ("ID_LITERAL_STRING");
  if (g == MMX_ALIAS) return mmx_symbol ("ID_ALIAS");
  if (g == MMX_CROSS) return mmx_symbol ("ID_CROSS");
  if (g == MMX_VCROSS) return mmx_symbol ("ID_VCROSS");
  if (g == MMX_TUPLETYPE) return mmx_symbol ("ID_TUPLETYPE");
  if (g == MMX_VTUPLETYPE) return mmx_symbol ("ID_VTUPLETYPE");
  if (g == MMX_FUNCTION) return mmx_symbol ("ID_FUNCTION");
  if (g == geninst ()) return symbol ("mmxt");
  int i= find_in_bindings (g, bindings);
  if (i != -1) return ext_tvar (i);
  return g;
}

list
extender::build_mmx_type (list l, list bindings) {
  if (nil (l)) return l;
  return cons (build_mmx_type (car (l), bindings),
	       build_mmx_type (cdr (l), bindings));
}

generic
extender::build_mmx_type (generic g, list bindings) {
  //cout << "b= " << bindings << lf;
  if (N(g) == 0) {
    generic id= build_mmx_type_sub (g, bindings);
    if (id != g) return id;
    int i= resolve_mmx_type (g, ex_types, 1, bindings);
    if (i != -1) return ext_id (i);
  }
  else {
    generic id= build_mmx_type_sub (g[0], bindings);
    if (id == g[0]) {
      int i= resolve_mmx_type (g, ex_types, 1, bindings);
      if (i != -1) id= ext_id (i);
    }
    if (id != g[0]) {
      generic mmxl= mmx_symbol ("list");
      generic l   = build_mmx_type (cdr (as_list (g)), bindings);
      return cons (mmxl, id, l);
    }
  }
  error (g, "Expected a Mathemagix type");
  // cerr << "Bindings were " << bindings << lf;
  return g;
}

/******************************************************************************
* Produce the type constructing routines
******************************************************************************/

void
extender::produce_glue_type (generic g, int i, bool prototype) {
  // cout << "Produce glue for type " << g << lf;
  if (!(is_func (g, MMX_DEFINE, 2) &&
	is_func (g[1], MMX_TYPE, 2) &&
	is_symbol (g[1][1]) &&
	is_symbol (g[2])))
    error (g, "Assertion failed in produce_glue_type");
  else {
    // the glue for the type
    generic id= ext_id (i);
    if (prototype) {
      generic vv= list (mmx_mirror, list (mmx_symbol ("symbol"), S(g[1][1])));
      generate (Assign (Type (id, Static (mmx_generic)), vv));
      return;
    }
    if (is_func (g[1][2], MMX_FUNCTION, 2)) {
      generic x  = symbol ("g");
      generic fun= ext_fun (i);
      generic arg= Type (x, mmx_generic);
      generic tp = Static (Function (arg, mmx_generic));
      generic l  = list (symbol ("cdr"), list (symbol ("as_list"), x));
      generic ret= list (symbol ("cons"), id, l);
      generate (Define (Type (fun, tp), Begin (Return (ret))));
      declare (S(g[1][1]), fun, build_mmx_type (g[1][2], list ()));
    }
    else declare (S(g[1][1]), id, build_mmx_type (g[1][2], list ()));

    // the corresponding conversion routine (mmx -> ccpp)
    if (!nil (gen_imports)) {
      generic ft = g[1][2];
      if (!is_func (ft, MMX_FUNCTION, 2))
	ft= Function (Cross (list ()), ft);
      generic x   = symbol ("g");
      generic fun = ext_conv (i);
      generic arg = Type (x, mmx_generic);
      generic tp  = Static (Function (arg, mmx_generic));
      list    orga= cdr (as_list (ft[1]));
      list    args= as_list (glue_convert_args (orga, x, table (), 1));
      generic op  = contype (g[2]);
      generic cvt = mmx_specialize ("glued", gentype, genid);
      generic ret = list (cvt, cons (op, args));
      generate (Define (Type (fun, tp), Begin (Return (ret))));

      generic decl= Dot (ext_cv (), symbol ("declare"));
      generic inst= list (decl, S (fun), id, fun);
      generate_init (inst);
    }
  }
}

void
extender::produce_glue_type (list l, int i, bool prototype) {
  if (!nil (l)) {
    produce_glue_type (car (l), i  , prototype);
    produce_glue_type (cdr (l), i+1, prototype);
  }
}

/******************************************************************************
* Deal with template arguments
******************************************************************************/

list
extender::glue_template_args (list l, int kind, int i, int n) {
  if (nil (l)) return l;
  generic g= car (l);
  list res;
  if (is_func (g, MMX_TYPE, 2)) {
    generic var = g[1];
    generic type= g[2];
    generic tp  = type_as_ccpp (type);
    generic org = Access (symbol ("g"), single (i));
    symbol  arg = string ("x_") * as_string (i);
    symbol  save= string ("save_") * as_string (n);
    switch (kind) {
    case 1: // retrieve template arguments
      if (type == MMX_CLASS) {
	generic sv= Assign (Type (save, gentype), curtype (n));
	generic nv= Assign (curtype (n), view_as_ccpp (org, type));
	res= list (sv, nv);
      }
      else res= Assign (Type (arg, tp), view_as_ccpp (org, type));
      break;
    case 2: // get new names of template arguments
      if (type == MMX_CLASS) res= list (generic (list (var, geninst (n))));
      else res= list (generic (list (var, arg)));
      break;
    case 3: // restore generic types
      if (type == MMX_CLASS) res= list (Assign (curtype (n), save));
      break;
    }
    if (type == MMX_CLASS) n++;
  }
  return res * glue_template_args (cdr (l), kind, i+1, n);
}

/******************************************************************************
* Deal with function names with specializations
******************************************************************************/

static generic
parse_angular (string s, int& i) {
  int start= i, n= N(s);
  for (; i<n; i++) {
    if ((s[i] == ',') || (s[i] == '>'))
      return symbol (s (start, i));
    if (s[i] == '<') break;
  }
  if (i == n) return symbol (s);
  list ret= generic (symbol (s (start, i)));
  while ((i<n) && (s[i] != '>')) {
    i++;
    ret= ret * list (parse_angular (s, i));
  }
  if (i<n) i++;
  return ret;
}

static generic
get_ccpp_funname (generic g, table t) {
  if (is_symbol (g) && is_string (S (as_symbol (g)))) {
    int     i= 0;
    string  s= as_string (S (as_symbol (g)));
    generic h= (starts (s, "operator ")? g: parse_angular (s, i));
    if (h == g) return ccpp_back (g);
    else {
      h= substitute (h, t);
      return cons (CCPP_SPECIALIZE, ccpp_back (car (h)), cdr (as_list (h)));
    }
  }
  else return g;
}

/******************************************************************************
* Produce the main routines
******************************************************************************/

list
extender::glue_convert_args (list l, generic x, table t, int i) {
  if (nil (l)) return l;
  generic gi = Access (x, single (i));
  generic cgi= view_as_ccpp (gi, car (l), t);
  return cons (cgi, glue_convert_args (cdr (l), x, t, i+1));
}

void
extender::produce_glue_routine (generic g, int i, list bindings) {
  // cerr << "Produce glue for routine " << g << lf;
  if (is_func (g, MMX_FORALL, 2))
    produce_glue_routine (g[2], i, bindings * list (g[1]));
  else if (!(is_func (g, MMX_DEFINE, 2) &&
	     is_func (g[1], MMX_TYPE, 2) &&
	     is_symbol (g[2]) &&
	     is_func (g[1][2], MMX_FUNCTION, 2) &&
	     is_func (g[1][2][1], MMX_CROSS)))
    error (g, "Assertion failed in produce_glue_routine");
  else {
    list code;
    generic x  = symbol ("g");
    generic fun= ext_fun (i);
    generic arg= Type (x, mmx_generic);
    generic tp = Static (Function (arg, mmx_generic));

    // Pretreatment of template arguments
    if (!nil (bindings)) {
      generic y = symbol ("h");
      generic ac= Assign (Type (y, mmx_generic), Access (x, single (0)));
      code= cons (ac, glue_template_args (bindings, 1));
      x= y;
    }
    table t= table_from_entries (glue_template_args (bindings, 2));

    // Execute main function and posttreatment of template arguments
    list    orga= cdr (as_list (g[1][2][1]));
    list    args= as_list (glue_convert_args (orga, x, t, 1));
    generic rtp = g[1][2][2];
    if (rtp == Cross (list ())) rtp= MMX_VOID;
    generic op= get_ccpp_funname (g[2], t);
    if (op == symbol ("new")) op= ccpp_type_protect (type_as_ccpp (rtp, t));
    generic ret= cons (op, args);
    if (rtp == MMX_VOID) {
      code= code * list (ret);
      code= code * glue_template_args (bindings, 3);
      code= code * list (Return (list (mmx_generic)));
    }
    else {
      ret= view_as_mmx (ret, rtp, t);
      if (nil (bindings)) code= code * list (Return (ret));
      else {
	generic r= symbol ("ret");
	code= code * list (Assign (Type (r, mmx_generic), ret));
	code= code * glue_template_args (bindings, 3);
	code= code * list (Return (r));
      }
    }
    generate (Define (Type (fun, tp), Begin (code)));

    // Temporary hack for tuples
    generic dtp= g[1][2];
    if (is_func (dtp, MMX_FUNCTION, 2))
      if (is_func (dtp[1], MMX_CROSS, 1))
	if (is_func (dtp[1][1], MMX_TUPLETYPE, 1))
	  dtp= Function (dtp[1][1], dtp[2]);

    // Put templates around declaration
    int j, n= N(bindings);
    dtp= build_mmx_type (dtp, bindings);
    for (j=n; j>0; j--) {
      generic mmxl= mmx_symbol ("list");
      dtp= list (mmxl, mmx_symbol ("ID_FORALL"), ext_tvar (j), dtp);
    }
    generic name= g[1][1];
    if (name == symbol ("new")) name= MMX_NEW (g[1][2][2]);
    declare (S(name), fun, dtp);
  }
}

void
extender::produce_glue_routine (list l, int i) {
  if (!nil (l)) {
    produce_glue_routine (car (l), i, list ());
    produce_glue_routine (cdr (l), i+1);
  }
}

/******************************************************************************
* Main routine for producing all glue
******************************************************************************/

void
extender::produce_glue () {
  int i;
  nr_ext_tvar= 0;
  generate_init (list (mmx_symbol ("register_type_id"), genid));
  produce_glue_type (ex_types, 1, false);
  produce_glue_routine (exports, N(ex_types) + 1);
  for (i=1; i<=nr_ext_tvar; i++) {
    generic var= ext_tvar (i);
    generic Ti = string ("T" * as_string (i));
    generic val= list (mmx_mirror, list (mmx_symbol ("symbol"), Ti));
    generate (Assign (Type (var, Static (mmx_generic)), val));
  }
  generic var= symbol ("init_" * the_prefix);
  generic tp = list (CCPP_EXTERN, string ("C"), Function (MMX_VOID, MMX_VOID));
  generic val= Begin (reverse (glue_init));
  generate (Define (Type (var, tp), val));
}

} // namespace mmx
