
/******************************************************************************
* MODULE     : glued.cc
* DESCRIPTION: Externally glued objects
* COPYRIGHT  : (C) 2004  Bernard Mourrain and Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __MMX_GLUED_HH
#define __MMX_GLUED_HH
#include <config.hh>
#include <generic.hh>
#include <list.hh>
#include <string.hh>
#include <gentable.hh>
#include <table.hh>
#include <symbol.hh>
#include <single.hh>
#include <printer.hh>
#include <literal.hh>
#include <alias.hh>
#include <mirror.hh>
#include <value.hh>
#include <function.hh>
#include <symbols.hh>
namespace mmx {

typedef generic (*mmx_function) (generic);
void declare (symbol var_name, generic val, generic type);
void declare (symbol var_name, mmx_function fn, generic type);
void register_type_id (int type_id);
generic extern_eval (generic g);
// defined in extern.cc and evaluate.cc

/******************************************************************************
* The glued template class
******************************************************************************/

template<typename T, int I> struct glued;
template<typename T, int I> ostream& operator << (ostream&, glued<T,I>);
template<typename T, int I> glued<T,I> as_glued (generic g);

template<typename T, int I>
struct glued_rep: generic_rep {
public:
  T& data; // the glued data (which can also be an alias)

protected:
  int Type ();
  void Print (ostream& out);
  // int Hash ();
  // bool Equal (generic g) const;
  // generic Copy ();

public:
  inline glued_rep (T& r): data (r) {}

  friend struct glued<T,I>;
  friend ostream& operator << <> (ostream& out, glued<T,I> a);
  friend glued<T,I> as_glued<T,I> (generic g);
};

template<typename T, int I>
struct glued {
protected:
  glued_rep<T,I>* rep;
  inline glued<T,I> (glued_rep<T,I>* r): rep (r) {}

public:
  inline glued<T,I> (T r): rep (new glued_rep<T,I> (*new T (r))) {}
  inline glued<T,I> (T* r): rep (new glued_rep<T,I> (*r)) {}
  inline operator generic () { return generic (rep); }
  inline glued_rep<T,I>* operator -> () { return rep; }
  inline bool operator == (glued<T,I> g) { return (rep->data == g->data); }
  inline bool operator != (glued<T,I> g) { return (rep->data != g->data); }

  friend struct glued_rep<T,I>;
  friend ostream& operator << <> (ostream& out, glued<T,I> x);
  friend glued<T,I> as_glued<T,I> (generic g);
};

/******************************************************************************
* Basic routines for glued template class
******************************************************************************/

template<typename T, int I> int
glued_rep<T,I>::Type () {
  return I;
}

template<typename T, int I> void
glued_rep<T,I>::Print (ostream& out) {
  out << glued<T,I> (this);
}

template<typename T, int I> ostream&
operator << (ostream& out, glued<T,I> x) {
  return out << "<glued object>";
  // return out << x->data;
}

template<typename T, int I> inline glued<T,I>
as_glued (generic g) {
  if (gentype(g) != I) type_mismatch (g, "glued");
  return glued<T,I> ((glued_rep<T,I>*) g.rep); 
}

/******************************************************************************
* Main converters between glue and generics
******************************************************************************/

template<class T, int I> struct exporter {
  static inline generic
  helper (T x) {
    return glued<T,I> (x);
  }
};

template<class T, int I> struct importer {
  static inline T
  helper (generic g) {
    return as_glued<T,I>(g)->data;
  }
};

template<class T, int I> struct exporter<T&, I> {
  static inline generic
  helper (T& x) {
    generic* ptr= (generic*) new glued<T,I> (&x);
    return alias (*ptr);
  }
};

template<class T, int I> struct importer<T&, I> {
  static inline T&
  helper (generic g) {
    return as_glued<T,I> (as_genalias (g) -> get ()) -> data;
  }
};

template<class T, int I> generic
as_mmx (T x) {
  return exporter<T, I>::helper (x);
}

template<class T, int I> T
as_cpp (generic g) {
  return importer<T, I>::helper (g);
}

/******************************************************************************
* Converters for built-in Mathemagix types
******************************************************************************/

template<int I> struct exporter<int, I> {
  static inline generic
  helper (int x) {
    return single (x);
  }
};

template<int I> struct importer<int, I> {
  static inline int
  helper (generic g) {
    return as_int (g);
  }
};

template<int I> struct exporter<bool, I> {
  static inline generic
  helper (bool x) {
    return single (x);
  }
};

template<int I> struct importer<bool, I> {
  static inline bool
  helper (generic g) {
    return as_bool (g);
  }
};

template<int I> struct exporter<ostream&, I> {
  static inline generic
  helper (ostream& x) {
    return printer (x);
  }
};

template<int I> struct importer<ostream&, I> {
  static inline ostream&
  helper (generic g) {
    return as_ostream (g);
  }
};

template<int I> struct exporter<int&, I> {
  static inline generic
  helper (int& x) {
    generic* ptr= (generic*) new single (&x);
    return alias (*ptr);
  }
};

template<int I> struct importer<int&, I> {
  static inline int&
  helper (generic g) {
    return as_int (as_genalias (g) -> get ());
  }
};

template<int I> struct exporter<bool&, I> {
  static inline generic
  helper (bool& x) {
    generic* ptr= (generic*) new single (&x);
    return alias (*ptr);
  }
};

template<int I> struct importer<bool&, I> {
  static inline bool&
  helper (generic g) {
    return as_bool (as_genalias (g) -> get ());
  }
};

/******************************************************************************
* Conversions from and to geninst structures
******************************************************************************/

#define GENINST_CONVERTERS(T)						\
namespace mmx {								\
  template<int I> struct exporter<T, I> {				\
    static inline generic						\
    helper (T x) {							\
      return *((generic*) x.rep);					\
    }									\
  };									\
									\
  template<int I> struct importer<T, I> {				\
    static inline T							\
    helper (generic g) {						\
      if (gentype(g) == I) {						\
	cerr << "\nI= " << I << "\n";					\
	fatal_error ("Type mismatch", "importer::helper", "glued.hh");	\
      }									\
      return T ((void*) new generic (g));				\
    }									\
  };									\
									\
  template<int I> struct exporter<T&, I> {				\
    static inline generic						\
    helper (T& x) {							\
      return alias (*((generic*) x.rep));				\
    }									\
  };									\
									\
  template<int I> struct importer<T&, I> {				\
    static inline T&							\
    helper (generic g) {						\
      generic& ref= as_genalias (g) -> get_ref ();			\
      if (gentype(ref) == I) {						\
	cerr << "\nI= " << I << "\n";					\
	fatal_error ("Type mismatch", "importer::helper", "glued.hh");	\
      }									\
      T* ptr= new T ((void*) (&ref));					\
      return *ptr;							\
    }									\
  };									\
}

/******************************************************************************
* Conversions from and to geninst_nr structures
******************************************************************************/

#define GENINSTX_CONVERTERS(T,C)					  \
namespace mmx {								  \
  template<int I> struct exporter<T, I> {				  \
    static generic							  \
    helper (T x) {							  \
      if (C->internal()) return glued<int,I> ((int*) x.rep);		  \
      else return *((generic*) x.rep);					  \
    }									  \
  };									  \
									  \
  template<int I> struct importer<T, I> {				  \
    static T								  \
    helper (generic g) {						  \
      if (C->internal()) return T ((void*) &as_glued<int,I> (g) -> data); \
      else return T ((void*) new generic (g));				  \
    }									  \
  };									  \
									  \
  template<int I> struct exporter<T&, I> {				  \
    static generic							  \
    helper (T& x) {							  \
      if (C->internal()) {						  \
	generic* ptr= (generic*) new glued<int,I> ((int*) x.rep);	  \
	return alias (*ptr);						  \
      }									  \
      else return alias (*((generic*) x.rep));				  \
    }									  \
  };									  \
									  \
  template<int I> struct importer<T&, I> {				  \
    static T&								  \
    helper (generic g) {						  \
      generic& ref= as_genalias (g) -> get_ref ();			  \
      if (C->internal()) {						  \
	T* ptr= new T ((void*) &as_glued<int,I> (ref) -> data);		  \
	return *ptr;							  \
      }									  \
      else {								  \
	T* ptr= new T ((void*) (&ref));					  \
	return *ptr;							  \
      }									  \
    }									  \
  };									  \
}

/******************************************************************************
* Conversions from and to gentype structures
******************************************************************************/

struct type_converter {
  table conv_routine;
  table conv_cache;

public:
  inline type_converter () {}
  inline void declare (symbol name, generic type_id, mmx_function fn);
  inline generic convert (generic g);
};

inline void
type_converter::declare (symbol name, generic type_id, mmx_function fn) {
  conv_routine (type_id)= function (name, fn);
}

inline generic
type_converter::convert (generic g) {
  if (conv_cache->contains (g)) return conv_cache [g];
  generic h= (is_list (g)? g: generic (list (g)));
  if (!conv_routine->contains (car (h))) {
    conv_cache (g)= generic ();
    return generic ();
  }
  function fn= as_function (conv_routine (car (h)));
  generic ret= fn (g);
  conv_cache (g)= ret;
  return ret;
}

// FIXME: references to classes not implemented

#define GENTYPE_CONVERTERS(T,R)			\
namespace mmx {					\
  template<int I> struct exporter<T, I> {	\
    static inline generic			\
    helper (T x) {				\
      return x->mmxt;				\
    }						\
  };						\
						\
  template<int I> struct importer<T, I> {	\
    static T					\
    helper (generic g) {			\
      generic ret = CV.convert (g);		\
      if (nil (ret)) {				\
	R *ptr = new R ();			\
	ptr->mmxt = g;				\
	return T (ptr);				\
      }						\
      else return as_glued<T,I> (ret) -> data;	\
    }						\
  };						\
}

/******************************************************************************
* Glued tuples and vtuples
******************************************************************************/

// FIXME:
// we should rather import the access and size operators,
// postfix [] and operator #, from Tuple T

// ATTENTION:
// tuples  are internally stored as (tuple! item_1 ... item_n)
// vtuples are internally stored as (vtuple! tuple_1 ... tuple_r)

template<typename T, int I>
class glued_tuple {
  generic rep;

public:
  inline glued_tuple (generic g): rep (g) {}
  inline operator generic () { return rep; }
  inline T operator [] (int i) { return as_cpp<T,I> (rep[i+1]); }
};

template<typename T, int I> inline generic
as_generic (glued_tuple<T,I> g) {
  return (generic) g;
}

template<typename T, int I> inline int
N (glued_tuple<T,I> g) {
  return N ((generic) g) - 1;
}

template<typename T, int I>
class glued_tuple<glued_tuple<T,I>,I> {
  generic rep;

public:
  inline glued_tuple<glued_tuple<T,I>,I> (generic g): rep (g) {}
  inline operator generic () { return rep; }
  inline glued_tuple<T,I> operator [] (int i) {
    return glued_tuple<T,I> (rep[i+1]); }
};

template<typename T, int I> inline generic
as_generic (glued_tuple<glued_tuple<T,I>,I> g) {
  return (generic) g;
}

template<typename T, int I> inline int
N (glued_tuple<glued_tuple<T,I>,I> g) {
  return N ((generic) g) - 1;
}

} //namespace mmx

/******************************************************************************
* Some default templates
******************************************************************************/

template<typename T> std::ostream&
print_lisp (std::ostream& out, T x) {
  return out << x;
}

template<typename T> T
div (T x, T y) {
  return x/y;
}

template<typename T> T
mod (T x, T y) {
  return x%y;
}

#endif // __MMX_GLUED_HH
