
/******************************************************************************
* MODULE     : generic.cc
* DESCRIPTION: Abstract types
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#include <generic.hh>
#include <list.hh>
#include <table.hh>
#include <single.hh>
#include <symbol.hh>
#include <mirror.hh>
#include <literal.hh>
#include <genalias.hh>
#include <value.hh>
namespace mmx {

/******************************************************************************
* Basic routines
******************************************************************************/

int
generic_rep::Type () {
  return UNKNOWN_TYPE;
}

int
generic_rep::Hash () {
  return ((int) this) >> 2;
}

bool
generic_rep::Equal (generic g) {
  return ((void*) this) == ((void*) g.rep);
}

generic
generic_rep::Copy () {
  return generic (this);
}

int
generic_rep::Arity () {
  return 0;
}

generic&
generic_rep::Access (int i) {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Access", "generic.cc");
  exit (1);
}

generic
generic_rep::Range (int begin, int end) {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no range function associated to generic",
	       "generic_rep::Range", "generic.cc");
  exit (1);
}

generic&
generic_rep::Car () {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no data access function associated to generic",
	       "generic_rep::Car", "generic.cc");
  exit (1);
}

generic
generic_rep::Cdr () {
  cout << lf << "g= " << generic (this) << lf;
  fatal_error ("no range function associated to generic",
	       "generic_rep::Cdr", "generic.cc");
  exit (1);
}

void
generic_rep::Print (ostream& out) {
  out << "#unknown";
}

/******************************************************************************
* Pritty printing
******************************************************************************/

static int cur_indent= 0;
static bool pritty_print = false;
static char indent_s[]  = "| ";
static char unindent_s[]= "\b\b";
static char hrule_s[]= "-----------------------------------------------------------------------------\n";

ostream&
operator << (ostream& out, print_format x) {
  int i;
  switch (x) {
  case blank:
    indent_s[0]= ' ';
    indent_s[1]= ' ';
    break;
  case stroke:
    indent_s[0]= '|';
    indent_s[1]= ' ';
    break;
  case indent:
    cur_indent++;
    out << indent_s;
    break;
  case unindent:
    cur_indent--;
    out << unindent_s;
    break;
  case lf:
    out << "\n";
    for (i=0; i<cur_indent; i++)
      out << indent_s;
    break;
  case pritty:
    pritty_print = true;
    break;
  case normal:
    pritty_print = false;
    break;
  case hrule:
    out << hrule_s;
    break;
  }
  return out;
}

void
print_entries (ostream& out, generic l, bool flag) {
  int n= N(l);
  if ((n == 0) && (!nil (l))) out << "{" << l << "}";
  else {
    out << (flag? "(": "<");
    if (pritty_print) {
      if (n != 0) {
	if (N(car(l)) == 0) {
	  out << car (l);
	  l= cdr (l); n--;
	}
	cur_indent++;
	while (n != 0) {
	  out << lf << car (l);
	  l= cdr (l); n--;
	}
	cur_indent--;
      }
    }
    else while (n != 0) {
      out << car (l);
      l= cdr (l); n--;
      if (n != 0) out << " ";
    }
    out << (flag? ")": ">");
  }
}

/******************************************************************************
* Locations in source code
******************************************************************************/

static table source_location;
static void locate_sub (generic g, int& begin, int& end);

void
generic_rep::index (int begin, int end) {
  single address ((int) ((void*) this));
  source_location (address)= list (single (begin), single (end));
}

void
generic_rep::index (generic g) {
  int begin, end;
  locate (g, begin, end);
  index (begin, end);
}

generic
extend (generic g, generic with) {
  if (nil (g) || nil (with)) return g;
  int begin=  0x40000000;
  int end  = -0x40000000;
  locate_sub (g, begin, end);
  if ((begin != 0x40000000) && (end != -0x40000000))
    g -> index (begin, end);
  return g;
}

static void
locate_sub (generic g, int& begin, int& end) {
  if (nil (g)) return;
  single address ((int) ((void*) g.rep));
  if (source_location -> contains (address)) {
    generic location= source_location [address];
    begin= min (begin, as_int (location[0]));
    end  = max (end  , as_int (location[1]));
  }
  else switch (gentype (g)) {
  case SYMBOL_TYPE:
    locate_sub (S (g), begin, end);
    break;
  case MIRROR_TYPE:
    locate_sub (S (as_mirror (g)), begin, end);
    break;
  case ALIAS_TYPE:
  case ALIASES_TYPE:
    locate_sub (as_genalias (g) -> get (), begin, end);
    break;
  case LIST_TYPE:
    locate_sub (car (g), begin, end);
    locate_sub (cdr (g), begin, end);
    break;
  case VALUE_TYPE:
    locate_sub (V (g), begin, end);
    break;
  case LITERAL_INT_TYPE:
  case LITERAL_FLOAT_TYPE:
  case LITERAL_STRING_TYPE:
    locate_sub (S (as_literal (g)), begin, end);
    break;
  }
}

void
locate (generic g, int& begin, int& end) {
  begin=  0x40000000;
  end  = -0x40000000;
  locate_sub (g, begin, end);
  if (begin ==  0x40000000) begin= -1;
  if (end   == -0x40000000) end  = -1;
}

int
generic_rep::begin () {
  int _begin, _end;
  locate (this, _begin, _end);
  return _begin;
}

int
generic_rep::end () {
  int _begin, _end;
  locate (this, _begin, _end);
  return _end;
}

/******************************************************************************
* Other routines
******************************************************************************/

generic&
S (generic g) {
  if (gentype (g) == SYMBOL_TYPE) return S (as_symbol (g));
  type_mismatch (g, "symbol", "S", "generic.cc");
  exit (1);
}

void
fatal_error (char* message, char* routine, char* file) {
  cerr << "\nFatal error: " << message
       << "\n             in '" << routine
       << "' in file '" << file << "'\n";
#ifdef DEBUG_ON
  cerr << 1/0 << "\n";
#endif
  exit (1);
}

void
type_mismatch (generic g, char* expected_type) {
  cerr << "\nFatal error: expected a value of type " << expected_type
       << "\n             in 'as_" << expected_type
       << "' in file '" << expected_type << ".hh'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
#ifdef DEBUG_ON
  cerr << 1/0 << "\n";
#endif
  exit (1);
}

void
type_mismatch (generic g, char* expected_type, char* routine, char* file) {
  cerr << "\nFatal error: expected a value of type " << expected_type
       << "\n             in '" << routine
       << "' in file '" << file << "'"
       << "\nValue was  : " << g
       << "\nType was   : " << gentype (g) << "\n";
#ifdef DEBUG_ON
  cerr << 1/0 << "\n";
#endif
  exit (1);  
}

} // namespace mmx
