
/******************************************************************************
* MODULE     : string.hh
* DESCRIPTION: Fixed size strings. Zero-characters are allowed in strings.
* COPYRIGHT  : (C) 2000  Joris van der Hoeven
*******************************************************************************
* This software falls under the GNU general public license and comes WITHOUT
* ANY WARRANTY WHATSOEVER. See the file $TEXMACS_PATH/LICENSE for more details.
* If you don't have this file, write to the Free Software Foundation, Inc.,
* 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
******************************************************************************/

#ifndef __STRING_HH
#define __STRING_HH
#include <generic.hh>
namespace mmx {

class string_rep: public generic_rep {
private:
  int n;
  char* a;

  int Type ();
  int Hash ();
  bool Equal (generic g);
  generic Copy ();
  void Print (ostream& out);

public:
  inline string_rep (): n(0), a(NULL) {}
         string_rep (int n);
  inline ~string_rep () { if (n!=0) delete[] a; }
  void resize (int n2);

  friend class string;
  friend inline int N (string s);
  friend inline char* S (string s);
};

class string {
private:
  string_rep* rep;
  inline string (string_rep* rep2): rep (rep2) {}

public:
  inline string (): rep (new string_rep()) {}
  inline string (int n): rep (new string_rep (n)) {}
  string (char c);
  string (char *s);
  string (char *s, int n);
  inline friend string as_string (generic g) {
    if (gentype(g) != STRING_TYPE) type_mismatch (g, "string");
    return string ((string_rep*) g.rep); }
  inline operator generic () { return generic (rep); }
  inline string_rep* operator -> () { return rep; }

  inline friend int N (string s) { return s.rep->n; }
  inline friend char* S (string s) { return s.rep->a; }
  inline char& operator [] (int i) { return rep->a[i]; }
  string operator () (int start, int end);
  bool operator == (char* s);
  bool operator != (char* s);
  bool operator == (string s);
  bool operator != (string s);
  friend string copy (string s);
  friend string as_string (int i);
  friend char*  as_charp (string s);
  friend ostream& operator << (ostream& out, string s);
  friend void raw_print (ostream& out, string s);

  friend class string_rep;
};

string& operator << (string& s, char c);
string& operator << (string& s, string t);
string& operator << (string& s, int i);
string  operator *  (string s, string t);
bool    operator <  (string s, string t);
bool    operator <= (string s, string t);

string  escape   (string s);
string  unescape (string s);
string  quote    (string s);
string  unquote  (string s);
string  delsimpl (string s);
string  locase   (string s);

bool    starts (string s, string what);
bool    ends (string s, string what);

int     as_int    (string s);
double  as_double (string s);

}
#endif // __STRING_HH
